"
 COPYRIGHT (c) 1995 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

AbstractSourceCodeManager subclass:#CVSSourceCodeManager
	instanceVariableNames:''
	classVariableNames:'Verbose CVSRoot CVSWorkDirectory RemoteCVS'
	poolDictionaries:''
	category:'System-SourceCodeManagement'
!

!CVSSourceCodeManager class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1995 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    SourceCodeManager which accesses the sourcecode through cvs.
    It requires the shell environment variable CVSROOT to be set
    to the top of the repository tree. (or alternatively, the CVSRoot
    variable being set by a private.rc file).
    Under that top, for each module (especially the stx module itself), 
    a subdirectory containing that modules directories must exist.

    If not specified otherwise, the module defaults to 'stx'. 
    If a class has its home in another module, it has to be compiled
    with a corresponding entry in the package string (see below).
    Within a module, classes are organized in classlibraries, and the corresponding
    source code is found in various subdirectories of CVSROOT. The directories
    name can also be provided by the package string - if not, it defaults to the
    libraries name.
    The package string controls all this, and is REQUIRED in order for the sourceCodeManager
    to be able to retrieve a classes source.
    The package strings format is:
        '...any infotext....(sourceCodeInfo)'
    the sourceCOdeInfo consists of multiple entries, separated by colons.
    The following formats are allowed:
        (libraryName)                   - module defaults to 'stx'; directory to the name of the library
        (foo/bar/x)                     - module defaults to 'stx'; dir is 'foo/bar/x'; library name defaults to x
        (module:directory:libname)      - specifies all components
    The first formats are intermediate, for backward compatibility. We urge everyone
    to use the last format (by changing the Make.proto files and add a -P flag to
    the stc compiler flags (see libbasic/Make.proto as an example).

    Notice, that the sourceCodeManager is (currently) only consulted, if no
    source subdirectory exists, or no source file is found there. If it does exist,
    that one is supposed to contain an up-to-date version of the classes source
    (this is temporary - in the future the probe order will be reversed, trying
     the source repository first).
    The source directory is required for systems which do NO source code control.

    Setup:
        - make certain, that the commands 'cvs' and 'co' are available on your system.
        - add setup for the CVSROOT environment variable to your .bashrc / .login / .profile
        - Make certain that either no local source directory exists, or it is empty or it contains
          only source code for files NOT found in the repository (this will not be required in future versions).

    Caveat:
        currently, the rcs container is accessed directly, since cvs mangles the files ident-string and
        thus does not correctly recreate the original file (all source is offset by some characters w.r.t. the
        compiled file).
        Therefore, sevrer-mode CVS is not supported.

    [author:]
        Claus Gittinger

    [see also:]
        ( cvs manual page :man: cvs )
        ( Coding style used in Smalltalk/X classes :html: programming/codingStyle.html )

    [restrictions:]
        commercial version only
"
! !

!CVSSourceCodeManager class methodsFor:'initialization'!

initialize
    "check if $CVSROOT (the shell variable) or CVSRoot (the smalltalk global)
     is set, check if cvs commands can be executed and install myself as the
     SourceCodeManager if this works out ok"

    |s f top|

    "/
    "/ if CVSROOT is non-empty and both cvs and co are available
    "/ as commands, assume this system uses a cvs source code management.
    "/
    CVSRoot isNil ifTrue:[
        CVSRoot := OperatingSystem getEnvironment:'CVSROOT'
    ].

    CVSRoot isNil ifTrue:[
        'SourceCodemanager [info]: disabled since $CVSROOT is not set' infoPrintCR.
        ^ self
    ].

    (OperatingSystem canExecuteCommand:'cvs') ifFalse:[
        'SourceCodemanager [warning]: disabled since no >> cvs << command found' infoPrintCR.
        ^ self
    ].
    (OperatingSystem canExecuteCommand:'co') ifFalse:[
        'SourceCodemanager [warning]: disabled since no >> co << command found' infoPrintCR.
        ^ self
    ].
    (OperatingSystem canExecuteCommand:'rlog') ifFalse:[
        'SourceCodemanager [warning]: disabled since no >> rlog << command found' infoPrintCR.
        ^ self
    ].

    RemoteCVS := true.
    UseWorkTree := false.

    self initCacheDirPath.
    CachingSources := true.

    "/
    "/ check if there is an stx directory there
    "/
    ((f := CVSRoot asFilename) exists 
    and:[f isDirectory
    and:[(top := f construct:'stx') exists
    and:[top isDirectory
    and:[top isReadable]]]]) ifTrue:[
        RemoteCVS := false
    ].

    Smalltalk at:#SourceCodeManager put:self.
    DefaultManager := self.    

    ('SourceCodemanager [info]: repository CVSROOT is ''' , CVSRoot , '''.') infoPrintCR.

    "/
    "/ optionally set the WorkTreeDirectoryName from $STX_WORKTREE;
    "/ if non-nil, a working tree is kept there
    "/ and updated/commited files are not removed.
    "/ If you use a regular (make-) tree, 
    "/ set WorkTreeDirectoryName (or the environment variable) to that.
    "/
    "/ this is not yet finished.
    "/
    s := OperatingSystem getEnvironment:'STX_WORKTREE'.
    s notNil ifTrue:[
        WorkTreeDirectoryName := s.
        UseWorkTree := true.
    ]

    "
     AbstractSourceCodeManager initialize
     CVSSourceCodeManager initialize
    "

    "
     CVSRoot := '/files/CVS'.
     AbstractSourceCodeManager initialize.
     CVSSourceCodeManager initialize.

    "

    "Created: 4.11.1995 / 19:14:38 / cg"
    "Modified: 19.12.1995 / 14:25:46 / stefan"
    "Modified: 10.1.1997 / 15:30:16 / cg"
! !

!CVSSourceCodeManager class methodsFor:'accessing'!

cvsRootDirectory:aDirectoryName
    "set the CVSROOT directory (i.e. the repositories name).
     Can be used from an rc-script, to override the CVSROOT shell 
     variable setting."

    CVSRoot := aDirectoryName.

    "Created: 9.9.1996 / 16:48:59 / cg"
    "Modified: 12.9.1996 / 02:37:30 / cg"
!

repositoryName
    "return the name of the repositor"

    ^ CVSRoot

    "Modified: 12.9.1996 / 02:37:30 / cg"
    "Created: 14.9.1996 / 13:22:05 / cg"
!

repositoryName:aDirectoryName
    "set the name of the repositor"

    self cvsRootDirectory:aDirectoryName

    "Modified: 12.9.1996 / 02:37:30 / cg"
    "Created: 14.9.1996 / 13:22:24 / cg"
! !

!CVSSourceCodeManager class methodsFor:'private'!

checkinTroubleDialog:title message:message log:log abortable:abortable option:optionTitle
    "trouble checking in - open a dialog"

    ^ self
        checkinTroubleDialog:title 
        message:message 
        log:log 
        abortable:abortable 
        option:optionTitle 
        option2:nil

    "Created: 10.12.1995 / 17:34:33 / cg"
    "Modified: 12.9.1996 / 02:39:06 / cg"
!

checkinTroubleDialog:title message:message log:log abortable:abortable option:optionTitle option2:optionTitle2
    "trouble checking in - open a dialog"

    |l box list listView optionPressed option2Pressed|

    l := log collect:[:line | line withTabsExpanded].
    list := SelectionInList with:l.

    box := Dialog new.
    box label:(title).

    (box addTextLabel:message) borderWidth:0.

    listView := SelectionInListView on:list.
    listView disable.
    listView height:200.
    box addComponent:(HVScrollableView forView:listView miniScrollerH:true) tabable:true.
    box addVerticalSpace.

    abortable ifTrue:[
        box addAbortButton
    ].
    optionTitle notNil ifTrue:[
        box addOkButton:(Button label:optionTitle action:[optionPressed := true. box hide]).
    ].
    optionTitle2 notNil ifTrue:[
        box addOkButton:(Button label:optionTitle2 action:[option2Pressed := true. box hide]).
    ].
    box addOkButton.

    box extent:(box preferredExtent).
    box minExtent:box extent.
    box maxExtent:box extent.

    box showAtPointer.

    box destroy.
    optionPressed == true ifTrue:[^ #option].
    option2Pressed == true ifTrue:[^ #option2].
    ^ box accepted

    "Created: 9.9.1996 / 19:12:45 / cg"
    "Modified: 12.9.1996 / 02:39:10 / cg"
!

createEntryFor:pathInRepository in:aDirectory  revision:rev date:date special:special overwrite:overwrite
    "create an CVS/Repository and CVS/Entries entry for filename. 
     If the path does not exist, it will be created
     If overwrite is true, an existing Entry will be overwritten.
    "

    |filename cvsDir name entries firstPart s|

    filename := pathInRepository asFilename.
    cvsDir := (aDirectory asFilename construct:'CVS') recursiveMakeDirectory.
    entries := cvsDir construct:'Entries'.
    name := filename baseName.
    firstPart := '/', name, '/'.

    s := entries readStream.
    s isNil ifTrue:[
	"/
	"/ no Repository yet, create one
	"/
	|index root|

	index := CVSRoot indexOf:$:.
	index > 0 ifTrue:[
	    root := CVSRoot copyFrom:(index+1).
	] ifFalse:[
	    root := CVSRoot.
	].
	s := (cvsDir construct:'Repository') writeStream.
	s nextPutLine:(root, '/', filename directory asString).
	s close.

	"/ make new Entries file.
	s := entries writeStream.
    ] ifFalse:[
	|newName newStream line|

	newName := Filename newTemporaryIn:cvsDir.
	newStream := newName writeStream.
	[(line := s nextLine) notNil] whileTrue:[
	    (line findString:firstPart) ~= 1 ifTrue:[
		"/ we have found an entry for ourself.
		"/ If we do not want to overwrite it, clean up everything and return.
		overwrite ifFalse:[
		    newName remove.
		    newStream close.
		    s close.
		    ^ self
		].
		newStream nextPutLine:line
	    ].
	].
	newName renameTo:entries.
	s close.
	s := newStream.
    ].
    s nextPutLine:firstPart, rev, '/', date, '/', special, '/'.
    s close.

    "
	self createEntryFor:'foo/xxx/yyy' in:'foo/xxx' revision:'1.22' date:'dummy' special:''
	self createEntryFor:'foo/xxx/yyy' in:'foo/xxx' revision:'3.44' date:'new dummy' special:'special'
    "

    "Modified: 19.12.1995 / 22:02:56 / stefan"
    "Created: 20.12.1995 / 11:09:54 / stefan"
!

createTempDirectory:packageDir forModule:moduleDir
    "create a temp directory for checking out"

    |tempdir dir|

    tempdir := Filename newTemporaryIn:nil.
    tempdir exists ifTrue:[
        tempdir recursiveRemove.
    ].
    tempdir makeDirectory.

    moduleDir notNil ifTrue:[
        dir := tempdir construct:moduleDir.
        dir makeDirectory.
        packageDir notNil ifTrue:[
            dir := dir construct:packageDir.
            dir recursiveMakeDirectory.
        ].
        dir exists ifFalse:[
            tempdir recursiveRemove.
            'SourceCodemanager [error]: cannot create temporary directory' errorPrintCR.
            ^ nil.
        ].
    ].
    ^ tempdir

    "self createTempDirectory:'fooPackage' forModule:'fooModule'"

    "Created: 9.12.1995 / 19:14:35 / cg"
    "Modified: 19.12.1995 / 16:13:02 / stefan"
    "Modified: 10.1.1997 / 15:29:14 / cg"
!

cvsTimeString:absoluteTime
    "convert an absoluteTime to a string suitable for the CVS/Entries file:

			 123456789012345678901234
			 Tue Dec 19 20:56:26 1995
    "

    |date time stream|

    date := absoluteTime asDate.
    time := absoluteTime asTime.
    stream := WriteStream on:(String new:24).

    (#('Mon' 'Tue' 'Wed' 'Thu' 'Fri' 'Sat' 'Sun') at:(date dayInWeek)) printOn:stream.
    stream space.

    (#('Jan' 'Feb' 'Mar' 'Apr' 'May' 'Jun' 'Jul' 'Aug' 'Sep' 'Oct' 'Nov' 'Dec') 
	at:(date month)) printOn:stream.
    stream space.

    date day printOn:stream.
    stream space.

    time print24HourFormatOn: stream.
    stream space.

    date year printOn:stream.

    ^ stream contents


    "
     CVSSourceCodeManager cvsTimeString:(AbsoluteTime now)
    "

    "Modified: 20.12.1995 / 11:03:32 / stefan"
!

executeCommand:cvsCommand
    "execute command and prepend cvs command name and global options"

    ^ self executeCommand:cvsCommand inDirectory:nil

    "Created: 23.4.1996 / 14:33:02 / stefan"
!

executeCommand:cvsCommand inDirectory:dir
    "execute command and prepend cvs command name and global options.
     if dir ~= nil, execute command in that directory"

    |command|

    command := 'cvs -d ', CVSRoot, ' ', cvsCommand.
    dir notNil ifTrue:[
        command := 'cd ', dir, '; ', command.
    ].
    Verbose == true ifTrue:[
        ('CVSMGR: executing: ' , command) infoPrintCR.
    ].

    ^ OperatingSystem executeCommand:command

    "Created: 23.4.1996 / 13:42:47 / stefan"
    "Modified: 23.4.1996 / 15:24:00 / stefan"
    "Modified: 20.5.1996 / 10:31:33 / cg"
!

releaseAndRemove:tempdir outputTo:outputFilename
    "cleanup; release tree towards cvs and remove the temporary tree"

    |cmd redirect|

    outputFilename isNil ifTrue:[
        redirect := '/dev/null'
    ] ifFalse:[
        redirect := outputFilename name
    ].

    "/
    "/ release it towards cvs
    "/
    cmd := '-l release > ' , redirect.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
"/        'CVSMGR: failed to release: ' errorPrint. cmd errorPrintNL.
    ].

    tempdir recursiveRemove.

    "Created: 10.12.1995 / 12:56:15 / cg"
    "Modified: 23.4.1996 / 14:29:49 / stefan"
    "Modified: 12.9.1996 / 02:40:08 / cg"
!

repositoryTopDirectory
    "return the top of the repository. This handles remote CVS connections
     as well"

    RemoteCVS == true ifTrue:[
        |idx|

        idx := CVSRoot indexOf:$:.
        idx = 0 ifTrue:[
            ^ CVSRoot.
        ].
        ^ CVSRoot copyFrom:(idx+1).
    ].
    ^ CVSRoot

    "
     CVSSourceCodeManager repositoryTopDirectory
    "

    "Created: 25.11.1995 / 18:39:19 / cg"
    "Modified: 15.3.1996 / 15:46:27 / stefan"
    "Modified: 12.9.1996 / 02:40:38 / cg"
!

revisionStringFromLog:log entry:entry forClass:aClass
    "given a log entry, extract a revision string"

    |oldRev idx special|

    oldRev := aClass revisionString.
    special := ''.

    oldRev notNil ifTrue:[
        idx := oldRev lastIndexOf:$[.
        idx ~~ 0 ifTrue:[
            idx := oldRev indexOf:$[ startingAt:idx+1.
            idx ~~ 0 ifTrue:[
                special := ' ' , (oldRev copyFrom:idx).
            ]
        ].
    ].

    ^ '$Header: ' , (log at:#container) ,
      ' ' , (entry at:#revision) ,
      ' ' , (entry at:#date) , 
      ' ' , (entry at:#author) , 
      ' ' , (entry at:#state) ,
      ' $' ,
      special

    "Created: 16.9.1996 / 20:24:14 / cg"
    "Modified: 3.3.1997 / 22:31:51 / cg"
!

updatedRevisionStringOf:aClass forRevision:newRevision with:originalVersionString
    "update a revision string"

    |idx leftPart rightPart vsnString newVsn|

    "/ search for ,v
    idx := originalVersionString indexOfSubCollection:'.st,v'.
    idx == 0 ifTrue:[^ nil].

    leftPart := originalVersionString copyTo:(idx - 1 + 5).
    rightPart := (originalVersionString copyFrom:(idx + 5)) withoutSpaces.
    idx := rightPart indexOfSeparator.
    idx == 0 ifTrue:[^ nil].

    vsnString := rightPart copyTo:idx - 1.
    rightPart := rightPart copyFrom:idx + 1.

    newRevision isNil ifTrue:[
        (vsnString endsWith:'m') ifTrue:[
             "/ alread a modified class
"/            ('already modified: ' , vsnString) printNL.
            ^ nil
        ].
        newVsn := vsnString , 'm'
    ] ifFalse:[
        newVsn := newRevision
    ].

    ^ leftPart , ' ' , newVsn , ' ' , rightPart

    "Created: 7.12.1995 / 20:23:38 / cg"
    "Modified: 16.9.1996 / 20:24:28 / cg"
! !

!CVSSourceCodeManager class methodsFor:'source code access'!

checkinClass:cls fileName:classFileName directory:packageDir module:moduleDir source:sourceFileName logMessage:logMessage force:force
    "enter a classes source code (which has been already filed out into sourceFileName)
     into the source repository. If the force argument is true, no merge is done;
     instead, the code is checked in as given (Dangerous).
     Return true if ok, false if not."

    |tempdir cmd checkoutName logMsg revision  
     cmdOut whatHappened s entry idx changeLog changesAsLogged l 
     newRevision newString binRevision className msg answer didMerge
     mySource mergedSource modulePath time 
     editor checkInRepaired didAccept emphasizedText repairedText out emSep comment|

    className := cls name.

    revision := cls revision.
    revision isNil ifTrue:[ 
        force ifFalse:[
            ('SourceCodeManager [warning]: class ' , className, ' has no revision string') errorPrintCR.
            (self confirm:('class ' , className, ' has no (usable) revision string.\\check in as newest ?') withCRs)
            ifFalse:[
                ^ false.
            ].
        ]
    ].
    (binRevision := cls binaryRevision) notNil ifTrue:[
        revision ~= binRevision ifTrue:[
            ('SourceCodeManager [info]: class ' , className , ' is based upon ' , binRevision , ' but has revision ' , revision) infoPrintCR
        ]
    ].

    revision isNil ifTrue:[
        revision := self newestRevisionOf:cls.
        revision isNil ifTrue:[
            force ifTrue:[
                revision := self newestRevisionInFile:classFileName directory:packageDir module:moduleDir.
            ] ifFalse:[
                revision := '1.0'
            ]
        ]
    ].

    logMsg := logMessage.
    (logMsg isNil or:[logMsg isEmpty]) ifTrue:[
        logMsg := 'checkin from browser'.
    ].

    cmdOut := Filename newTemporary.
    cmdOut exists ifTrue:[
        cmdOut remove.
    ].

    "/
    "/ in CVS, we have to checkout the file first, in order
    "/ to get up-to-date CVS entries, and also to be able to merge in
    "/ other users changes.
    "/

    "/
    "/ first, create a temporary work tree
    "/
    tempdir := self createTempDirectory:nil forModule:nil.
    tempdir isNil ifTrue:[^ false].

    "/
    "/ next, create CVS/Entries and CVS/Repository with version information of current version
    "/
    modulePath :=  moduleDir , '/' , packageDir. 
    checkoutName :=  modulePath , '/' , classFileName.

    "/
    "/ correct our current time, so that converting it will give us UTC
    "/
    time := AbsoluteTime now subtractSeconds:1.
    time := time addSeconds:(time utcOffset).

    self createEntryFor:checkoutName 
         in:(tempdir construct:modulePath) 
         revision:revision
         date:(self cvsTimeString:time)
         special:''
         overwrite:true.

    "/
    "/ copy-over our current version
    "/
    Object errorSignal handle:[:ex|
        tempdir recursiveRemove.
        'SourceCodeManager [error]: cannot copy-over filedOut class source' errorPrintCR.
        ^ false.
    ] do:[
        sourceFileName asFilename copyTo:(tempdir name , '/' , checkoutName).
    ].

    "/
    "/ synchronize i.e. merge in any changes
    "/
    self activityNotification:'merging ' , cls name , ' with repository version ...'.

    cmd := 'update ', checkoutName, ' > ', cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        force ifFalse:[
            'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.
            tempdir recursiveRemove.
            cmdOut remove.
            'SourceCodeManager [error]: cannot merge current source with repository version' errorPrintCR.
            ^ false.
        ].
    ] ifTrue:[
        "/
        "/ check what happened - the contents of the cmdOut file may be:
        "/   empty   -> nothing changed
        "/   M xxx   -> merged-in changes from other users
        "/   C xxx   -> a conflict occured and the differences have been merged into the source
        "/              needs special action
        "/
        cmdOut fileSize > 0 ifTrue:[
            whatHappened := cmdOut contentsOfEntireFile asString.
        ].
    ].
    cmdOut remove.

    (whatHappened isNil or:[whatHappened isEmpty]) ifTrue:[
        "/
        "/ no change
        "/
"/        Transcript showCR:'no change in ' , className , ' (repository unchanged)'.
        tempdir recursiveRemove.

        force ifFalse:[
            self information:'nothing changed in ' , className , ' (repository unchanged)'.
        ] ifTrue:[
            changeLog := self revisionLogInFile:classFileName directory:packageDir module:moduleDir.
            (changeLog isNil or:[(changeLog at:#revisions) size ~~ 1]) ifTrue:[
                'SourceCodeManager [error]: failed to update revisionString (no log)' errorPrintCR.
                cls updateVersionMethodFor:'$' , 'Header' , '$'.  "/ concatenated to avoid RCS expansion
            ] ifFalse:[
                entry := (changeLog at:#revisions) first.
                newString := self revisionStringFromLog:changeLog entry:entry forClass:cls.
                cls updateVersionMethodFor:newString.
                cls revision ~= newRevision ifTrue:[
                    'SourceCodeManager [error]: failed to update revisionString' errorPrintCR
                ]
            ]
        ].

        ^ true
    ].

    Verbose == true ifTrue:[
        ('CVSMGR: result is: ' , whatHappened) infoPrintCR.
    ].

    force ifFalse:[
        revision isNil ifTrue:[
            changeLog := self revisionLogOf:cls.
        ] ifFalse:[
            changeLog := self revisionLogOf:cls fromRevision:(self revisionAfter:revision) toRevision:nil.
        ].
        s := WriteStream on:String new.
        self writeRevisionLogMessagesFrom:changeLog withHeader:false to:s.
        changesAsLogged := s contents.
    ].

    didMerge := false.

    (force or:[whatHappened startsWith:'M ']) ifTrue:[
        "/
        "/ merged in changes
        "/
        (force or:[(changeLog at:#revisions) isEmpty]) ifTrue:[
            "/
            "/ pretty good - nothing has changed in the meanwhile
            "/
            Transcript showCR:'checking in ' , className , ' ...'
        ] ifFalse:[
            "/
            "/ someone else has changed things in the meanwhile, but there is no conflict
            "/ and version have been merged.
            "/
            didMerge := true.
            changesAsLogged := changesAsLogged asCollectionOfLines.

            s := WriteStream on:String new.
            cls fileOutOn:s withTimeStamp:false.
            mySource := s contents asString.
            mergedSource := (tempdir name , '/' , checkoutName) asFilename readStream contents asString.

            mySource = mergedSource ifTrue:[
                msg := 'The source of ' , className , ' has been changed in the meanwhile as listed below.

I have merged your version with the newest repository version, 
and found no differences between the result and your current version
(i.e. your version seemed up-to-date).'.

                self checkinTroubleDialog:'Merging versions'
                               message:msg 
                               log:changesAsLogged
                               abortable:false 
                               option:nil.
                didMerge := false.
            ] ifFalse:[
                msg := 'The source of ' , className , ' has been changed in the meanwhile as listed below.

If you continue, your new changes (based upon rev. ' , revision printString , ') will be MERGED
into the newest revision. This will combine the other version with your changes
into a new common revision which may be different from both. 
Although this is a nice feature, it may fail to create the expected result in certain situations. 

You should carefully check the result - by comparing the current version with the
most recent version in the repository. If that does not contain an acceptable version,
change methods as required and check in again. 
Be aware, that after that, the actual repository version is different from your current classes,
and you should update your class from the repository.

Continue ?'.

                answer := self checkinTroubleDialog:'Merging versions'
                               message:msg 
                               log:changesAsLogged
                               abortable:true
                               option:'stop - see first'.
                answer ~~ true ifTrue:[
                    answer == #option ifTrue:[
                        DiffTextView 
                            openOn:mySource
                            label:'current version'
                            and:mergedSource
                            label:'merged version'.
                            
                    ].
                    Transcript showCR:'checkin aborted - (no merge; repository unchanged)'.
                    tempdir recursiveRemove.
                    ^ false.
                ].
            ].

"/                changesAsLogged := (changesAsLogged asStringCollection collect:[:line | line withTabsExpanded]) asString.
"/                msg := 'The source of ' , className , ' has been changed in the meanwhile as follows:
"/' , changesAsLogged , '
"/
"/If you continue, your new changes (based upon rev. ' , revision , ') will be MERGED
"/into the newest revision. This will combine the other version with your changes
"/into a new common revision which is different from both. 
"/Although convenient, it may fail to create the expected result in certain situations. 
"/
"/You should carefully check the result - by comparing the current version with the
"/most recent version in the repository. If that does not contain an acceptable version,
"/change methods as required and check in again. Be aware, that the actual repository version
"/is different from your current classes.
"/
"/Continue ?'.
"/                (self confirm:msg) ifFalse:[
"/                    Transcript showCR:'checkin aborted - (no merge; repository unchanged)'.
"/                    tempdir recursiveRemove.
"/                    ^ false.
"/                ].
            Transcript showCR:'checking in ' , className , ' (merged other changes) ...'
        ]
    ] ifFalse:[
        (whatHappened startsWith:'C ') ifTrue:[
            "/
            "/ conflict; someone else checked in something in the meanwhile,
            "/ and there is a conflict between this version and the checked in version.
            "/

            changesAsLogged := changesAsLogged asCollectionOfLines.

            msg := 'The source of ' , className , ' has been changed in the meanwhile as listed below.

Your new changes (based upon rev. ' , revision printString , ') CONFLICT with those changes.

You should fix things by comparing your class with the most recent repository version
and change your methods avoiding conflicts. The checkin again.
'.

            answer := self checkinTroubleDialog:'Version conflict'
                 message:msg 
                 log:changesAsLogged
                 abortable:false
                 option:'show conflicts'
                 option2:'resolve conflicts'.

            answer == #option ifTrue:[
                "/
                "/ show conflicts in a 3-way DiffTextView ...
                "/
                Diff3TextView
                    openOnMergedText:(tempdir name , '/' , checkoutName) asFilename readStream contents 
                    label:'your version (checkin attempt)' 
                    label:'original (base version)' 
                    label:'newest repository version'. 
            ].

            checkInRepaired := false.
            answer == #option2 ifTrue:[
                "/
                "/ allow checkin of repair version
                "/ this is error prone ...
                "/
                "/
                "/ show merged version in an editor ...
                "/ ... accept will check it in.
                "/
                emphasizedText := (tempdir name , '/' , checkoutName) asFilename readStream contents.
                emSep := (Array with:(#color->Color black)
                             with:(#backgroundColor->Color green)).
                emphasizedText := Diff3TextView 
                            emphasizeMergedDiff3Text:emphasizedText 
                            emphasize1:(Array with:(#color->Color white)
                                              with:(#backgroundColor->Color blue))
                            emphasize2:(Array with:(#color->Color white)
                                              with:(#backgroundColor->Color red))
                            emphasizeSep:emSep.

                comment := 
'"/ ***************************************************************
"/ This text contains your current versions code (blue)
"/ merged with the conflicting code as found in the repository (red) which resulted
"/ from some other checkin.
"/ Each such conflict is surrounded by green text (like this paragraph).
"/ 
"/ Please have a look at ALL the conflicts and fix things as appropriate.
"/ Delete the green lines as a confirmation - I will not checkin the changed text,
"/ unless no more green parts are present. This includes this comment at the top.
"/ ***************************************************************
'.
                comment := (Text string:comment emphasis:emSep) asStringCollection.
                emphasizedText := comment , emphasizedText.

                didAccept := false. checkInRepaired := true.
                [didAccept not and:[checkInRepaired]] whileTrue:[
                    editor := EditTextView 
                                setupWith:emphasizedText
                                title:'Resolve conflicts in ' , className , ', then accept & close to checkin'.    

                    editor acceptAction:[:dummy |
                        repairedText := editor list.
                        didAccept := true.
                    ].
                    didAccept := false.
                    editor topView openModal.

                    didAccept ifFalse:[
                        (self confirm:'You did not accept the new text. Edit again ?')
                        ifFalse:[
                            checkInRepaired := false.
                        ]
                    ] ifTrue:[
                        "/ check if all green-stuff (separators) have been removed
                        (repairedText findFirst:[:line | line notEmpty and:[(line emphasisAt:1) = emSep]]) ~~ 0 ifTrue:[
                            self warn:'You have to look at ALL conflicts, and remove ALL green lines as a confirmation !!'.
                            didAccept := false.
                        ]
                    ].

                ].

                checkInRepaired ifTrue:[
                    out := (tempdir name , '/' , checkoutName) asFilename writeStream.
                    out isNil ifTrue:[
                        self warn:'could not write file ' , (tempdir name , '/' , checkoutName).
                        checkInRepaired := false.
                    ] ifFalse:[
                        out nextPutAll:(repairedText asString string).
                        didAccept := true.
                        out close.
                    ].
                ]
            ].

            checkInRepaired ifTrue:[
                Transcript showCR:'checking in ' , className , ' (manually repaired version) ...'
            ] ifFalse:[
                'SourceCodeManager [warning]: cannot (for now) checkin; conflicts found' infoPrintCR.
                Transcript showCR:'checkin of ' , className , ' aborted (conflicting changes; repository unchanged)'.
                tempdir recursiveRemove.
                ^ false.
            ]
        ] ifFalse:[
            "/
            "/ unexpected
            "/
            self warn:'unexpected message from CVS:
' , whatHappened , '

No checkin performed.'.
            Transcript showCR:'*** cannot checkin ' , className , ' (unexpected CVS response; repository unchanged)'.
            tempdir recursiveRemove.
            ^ false.
        ]
    ].


    "/
    "/ now check it in again
    "/
    self activityNotification:'saving ' , cls name , ' in repository ...'.

    logMsg := logMsg replChar:$"  withString:'\"'.
    cmd := 'commit -m "', logMsg, '" ', checkoutName, ' 2> ', cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut fileSize > 0 ifTrue:[
            whatHappened := cmdOut contentsOfEntireFile asString.
        ] ifFalse:[
            whatHappened := '<< no message >>'
        ].
        self warn:'The following problem was reported by cvs:

' , whatHappened , '

The class has NOT been checked into the repository.'.

       'SourceCodeManager [error]: cannot checkin modified class source' errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false.
    ].
    cmdOut fileSize > 0 ifTrue:[
        whatHappened := cmdOut contentsOfEntireFile asString.
    ] ifFalse:[
        whatHappened := nil
    ].

    tempdir recursiveRemove.
    cmdOut remove.

    "/
    "/ fetch the new revision nr as found in the commit commands output
    "/

    (whatHappened isNil or:[whatHappened isEmpty]) ifTrue:[
        'SourceCodeManager [error]: unexpected empty checkin command output' errorPrintCR.
    ] ifFalse:[
        whatHappened := whatHappened asCollectionOfLines asStringCollection.
        idx := whatHappened indexOfLineStartingWith:'new revision:'.
        idx == 0 ifTrue:[
            'SourceCodeManager [error]: unexpected checkin command output (no new-revision info)' errorPrintCR.
        ] ifFalse:[
            l := whatHappened at:idx.
            newRevision := (l copyFrom:14 to:(l indexOf:$; startingAt:14)-1) withoutSpaces.
        ]
    ].

    "/
    "/ if there was no merge (i.e. the current version has been checked in unchanged):
    "/   patch the classes revisionInfo (but keep binaryRevision unchanged) to the new revision
    "/   this makes everyone here believe, that the incore version of the class is based upon
    "/   the newly checked in version.
    "/   (however, the binaryRevision must remain as it is - we will need it to fetch the sourceCode
    "/    correctly for all unchanged methodss)
    "/
    "/ if there was a merge (i.e. the repository now contains a merge of the current and some
    "/ other version):
    "/    patch the classes revisionInfo (again, keep the binaryRevision) to the old revision
    "/    and add a 'm' (for merged).
    "/    If we later checkin again, the new checkin will be again based on the current revision
    "/
    newRevision notNil ifTrue:[
        didMerge ifFalse:[
            self activityNotification:'fetch new revision number of ', cls name.

            changeLog := self revisionLogOf:cls fromRevision:newRevision toRevision:newRevision.
            (changeLog isNil or:[(changeLog at:#revisions) size ~~ 1]) ifTrue:[
                force ifTrue:[
                    changeLog := self revisionLogInFile:classFileName directory:packageDir module:moduleDir.
                ].
            ].
            (changeLog isNil or:[(changeLog at:#revisions) size ~~ 1]) ifTrue:[
                'SourceCodeManager [error]: failed to update revisionString (no log)' errorPrintCR.
                cls updateVersionMethodFor:'$' , 'Header' , '$'.  "/ concatenated to avoid RCS expansion
            ] ifFalse:[
                entry := (changeLog at:#revisions) first.
                newString := self revisionStringFromLog:changeLog entry:entry forClass:cls.
                cls updateVersionMethodFor:newString.
                cls revision ~= newRevision ifTrue:[
                    'SourceCodeManager [error]: failed to update revisionString' errorPrintCR
                ]
            ]
        ] ifTrue:[
            newString := self updatedRevisionStringOf:cls forRevision:nil with:cls revisionString.
            cls updateVersionMethodFor:newString.
        ].
    ].

    Class addChangeRecordForClassCheckIn:cls.
    ^ true

    "
     SourceCodeManager checkinClass:Array logMessage:'testing only'
    "

    "Modified: 23.4.1996 / 14:44:12 / stefan"
    "Created: 11.9.1996 / 16:16:11 / cg"
    "Modified: 10.1.1997 / 17:30:59 / cg"
!

streamForClass:cls fileName:fileName revision:revision directory:packageDir module:moduleDir cache:cacheIt
    "extract a source file and return an open readStream on it,
     or nil if the extract failed. If revision is nil or (#newest), take
     the latest; otherwise, a specific revisions source is extracted.
     The cls argument is currently ignored, but may be used in future versions,
     to validate the correct container against the class."

    |dir cachedSourceFilename cachedFile cmd fullName revisionArg stream
     checkoutDir checkoutName fullTempName fullCachedName tempdir cmdOut
     classFileName|


    "/ if not already existing, create a cache directory
    "/ where we deposit sources.
    "/ this is used as a cache for further requests, since
    "/ accessing the repository may be slow.
    "/ (if cvs uses a remote connection via ppp, for example)
    "/

    (dir := self sourceCacheDirectory) isNil ifTrue:[
        'SourceCodeManager [warning]: no source cache directory' errorPrintCR.
    ].

    classFileName := fileName.
    (classFileName endsWith:',v') ifTrue:[
        classFileName := classFileName copyWithoutLast:2.
    ].
    (classFileName endsWith:'.st') ifTrue:[
        classFileName := classFileName copyWithoutLast:3.
    ].

    fullName := moduleDir , '/' , packageDir , '/' , classFileName , '.st'.

    (revision isNil or:[revision == #newest]) ifTrue:[
        cachedSourceFilename := classFileName.
    ] ifFalse:[
        cachedSourceFilename := classFileName , '_' , revision.
    ].

    dir notNil ifTrue:[
        cachedFile := (dir construct:cachedSourceFilename) asFilename.
        cachedFile exists ifTrue:[
            cachedFile fileSize < 10 ifTrue:[
                ('SourceCodeManager [warning]: existing: ', cachedFile name , ' seems corrupted.') errorPrintCR.
                cachedFile remove
            ] ifFalse:[
"/            ('SourceCodeManager [info]: found existing: ', cachedFile name) infoPrintCR.
                ^ cachedFile readStream
            ]
        ].
    ].

    "/
    "/ first, create a temporary work tree
    "/ Do not make module and package directories, their existence cause cvs checkout to fail in server mode
    "/
    tempdir := self createTempDirectory:nil forModule:nil.
    tempdir isNil ifTrue:[^ nil].

"/    tempdir := Filename newTemporaryIn:nil.
"/    tempdir exists ifTrue:[
"/        tempdir recursiveRemove.
"/    ].
"/    tempdir makeDirectory.
"/    dir := tempdir construct:moduleDir.
"/    dir makeDirectory.
"/    dir := dir construct:packageDir.
"/    dir recursiveMakeDirectory.
"/    dir exists ifFalse:[
"/        tempdir recursiveRemove.
"/        'CVSMGR: cannot create temporary directory' infoPrintCR.
"/        ^ nil.
"/    ].

    cmdOut := Filename newTemporary.
    cmdOut exists ifTrue:[
        cmdOut remove.
    ].

    "/
    "/ check it out there
    "/
    checkoutDir :=  moduleDir , '/' , packageDir.
    checkoutName :=  checkoutDir , '/' , classFileName , '.st'.

    (revision isNil or:[revision == #newest]) ifTrue:[
        cachedSourceFilename := classFileName.
        revisionArg := ''.
    ] ifFalse:[
        cachedSourceFilename := classFileName , '_' , revision.
        revisionArg := ' -r ' , revision.
    ].

    self activityNotification:'checking out source ' , checkoutName.
    cmd := '-l checkout' , revisionArg , ' ', checkoutName , ' > ' , cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ('SourceCodeManager [error]: cannot checkout ' , checkoutName) errorPrintCR.
        ^ nil.
    ].

    cmdOut remove.
    fullTempName := tempdir name , '/' , checkoutName.
    fullCachedName := CacheDirectoryName , '/' , cachedSourceFilename.

    cacheIt ifTrue:[
        (OperatingSystem errorSignal catch:[
            fullTempName asFilename moveTo:fullCachedName
        ]) ifTrue:[
            ('SourceCodeManager [error]: failed to rename ', fullTempName, ' to ', cachedSourceFilename) errorPrintCR.
            tempdir recursiveRemove.
            ^ nil
        ].
        stream := cachedFile readStream
    ] ifFalse:[
        stream := fullTempName asFilename readStream.
    ].

    self releaseAndRemove:tempdir outputTo:nil. 
    ^ stream

    "Created: 4.11.1995 / 19:46:20 / cg"
    "Modified: 23.4.1996 / 14:31:03 / stefan"
    "Modified: 10.1.1997 / 15:35:00 / cg"
! !

!CVSSourceCodeManager class methodsFor:'source code administration'!

checkForExistingContainerInModule:moduleDir package:packageDir container:fileName
    "check for a container to exist"

    |fullName tempdir ret|

    fullName := moduleDir , '/' , packageDir , '/' , fileName.
    RemoteCVS ifFalse:[
        (fullName endsWith:',v') ifFalse:[
            fullName := fullName , ',v'.
        ].
        ^ (CVSRoot , '/' , fullName) asFilename exists.
    ].

    (fullName endsWith:',v') ifTrue:[
        fullName := fullName copyWithoutLast:2.
    ].

    tempdir := Filename newTemporaryIn:nil.
    tempdir exists ifTrue:[
        tempdir recursiveRemove.
    ].
    tempdir makeDirectory.

    ret := self executeCommand:'-n checkout ', fullName inDirectory:tempdir pathName.
    tempdir recursiveRemove.
    ^ ret

    "
     CVSSourceCodeManager 
        checkForExistingContainerInModule:'stx' 
        package:'libbasic' 
        container:'Integer.st'     

     CVSSourceCodeManager 
        checkForExistingContainerInModule:'stx' 
        package:'libtool' 
        container:'AboutBox.st'  

     CVSSourceCodeManager 
        checkForExistingContainerInModule:'stx' 
        package:'libtool' 
        container:'AboutBox.st,v'     

     CVSSourceCodeManager 
        checkForExistingContainerInModule:'stx' 
        package:'libtool' 
        container:'FooBar.st'  

     CVSSourceCodeManager 
        checkForExistingContainerInModule:'DPU' 
        package:'test' 
        container:'marks'  
    "

    "Created: 9.12.1995 / 19:13:37 / cg"
    "Modified: 23.4.1996 / 12:53:49 / stefan"
    "Modified: 12.9.1996 / 02:41:20 / cg"
!

checkForExistingModule:moduleDir
    "check for a module to exist"

    |tempdir ret|

    RemoteCVS ifFalse:[
        ^ (CVSRoot , '/' , moduleDir) asFilename isDirectory.
    ].

    tempdir := Filename newTemporaryIn:nil.
    tempdir exists ifTrue:[
        tempdir recursiveRemove.
    ].
    tempdir makeDirectory.

    ret := self executeCommand:'-n checkout ', moduleDir inDirectory:tempdir pathName.
    tempdir recursiveRemove.
    ^ ret

    "
     CVSSourceCodeManager checkForExistingModule:'stx'
     CVSSourceCodeManager checkForExistingModule:'DPU'
     CVSSourceCodeManager checkForExistingModule:'cg'  
     CVSSourceCodeManager checkForExistingModule:'aeg' 
     CVSSourceCodeManager checkForExistingModule:'foo'   
    "

    "Created: 9.12.1995 / 19:13:37 / cg"
    "Modified: 23.4.1996 / 12:54:27 / stefan"
    "Modified: 12.9.1996 / 02:41:28 / cg"
!

checkForExistingModule:moduleDir package:packageDir
    "check for a package to exist"

    |fullName tempdir ret|

    fullName := moduleDir , '/' , packageDir.

    RemoteCVS ifFalse:[
        ^ (CVSRoot , '/' , fullName) asFilename isDirectory.
    ].

    tempdir := Filename newTemporaryIn:nil.
    tempdir exists ifTrue:[
        tempdir recursiveRemove.
    ].
    tempdir makeDirectory.

    ret := self executeCommand:'-n checkout ', fullName inDirectory:tempdir pathName.
    tempdir recursiveRemove.
    ^ ret

    "
     CVSSourceCodeManager checkForExistingModule:'stx' package:'libbasic' 
     CVSSourceCodeManager checkForExistingModule:'aeg' package:'libProgram'  
     CVSSourceCodeManager checkForExistingModule:'foo' package:'libbasic' 
     CVSSourceCodeManager checkForExistingModule:'foo' package:'bar'   
     CVSSourceCodeManager checkForExistingModule:'cg' package:'private'   
    "

    "Created: 9.12.1995 / 19:13:37 / cg"
    "Modified: 23.4.1996 / 13:37:33 / stefan"
    "Modified: 12.9.1996 / 02:41:39 / cg"
!

createContainerFor:aClass inModule:moduleDir package:packageDir container:fileName
    "create a container - this does an initial checkin
     (i.e. cvs add & cvs commit)"

    |fullName tempdir checkoutName cmdOut cmd tempFile idx aStream whatHappened l newRevision
     changeLog entry newString startIdx|

    "/
    "/ first, create a temporary work tree
    "/
    tempdir := self createTempDirectory:'' forModule:''.
    tempdir isNil ifTrue:[^ false].

    cmdOut := Filename newTemporary.
    cmdOut exists ifTrue:[
        cmdOut remove.
    ].

    fullName := moduleDir , '/' , packageDir , '/' , fileName.

    "/
    "/ first, check out everything there - this creates the CVS helpfiles
    "/ required later.
    "/
    checkoutName :=  moduleDir , '/' , packageDir.

    self activityNotification:'checking for directory ' , checkoutName.

    cmd := 'checkout -l ' , checkoutName , ' > ' , cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut remove.
        tempdir recursiveRemove.
        ('SourceCodeManager [error]: cannot checkout ' , checkoutName) errorPrintCR.
        ^ false.
    ].

    "/
    "/ create the source there
    "/
    tempFile := (tempdir name , '/' , checkoutName , '/' , fileName) asFilename.
    aStream := tempFile writeStream.
    aStream isNil ifTrue:[
        ('SourceCodeManager [error]: temporary fileout failed -> ', tempFile name) errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false
    ].

    Class fileOutErrorSignal handle:[:ex |
        'SourceCodeManager [error]: fileout failed' errorPrintCR.
        aStream close.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false
    ] do:[
        aClass fileOutOn:aStream withTimeStamp:false.
    ].
    aStream close.

    tempFile exists ifFalse:[
        'SourceCodeManager [error]: temporary fileout failed' errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false
    ].

    "/
    "/ and add it to the repository
    "/
    self activityNotification:'adding ' , fileName.

    cmd := 'add ' , fileName , ' > ' , cmdOut name.
    (self executeCommand:cmd inDirectory:(tempdir name , '/' , checkoutName)) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut remove.
        tempdir recursiveRemove.
        ('SourceCodeManager [error]: cannot checkout ' , checkoutName) errorPrintCR.
        ^ false.
    ].

    cmdOut remove.

    "/
    "/ commit
    "/
    self activityNotification:'comitting ' , fileName.

    cmd := 'commit -m "intitial checkin" 2> ' , cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        Verbose == true ifTrue:[
            'CVSMGR: failed to execute: ' infoPrint. cmd infoPrintCR.
        ].
        cmdOut fileSize > 0 ifTrue:[
            whatHappened := cmdOut contentsOfEntireFile asString.
        ] ifFalse:[
            whatHappened := '<< no message >>'
        ].
        self warn:'The following problem was reported by cvs:

' , whatHappened , '

The class has NOT been checked into the repository.'.

       'SourceCodeManager [error]: cannot checkin modified class source' errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false.
    ].
    whatHappened := cmdOut contentsOfEntireFile asString.

    "/
    "/ release it towards cvs
    "/
    self releaseAndRemove:tempdir outputTo:nil. 

    "/
    "/ good - its in the CVS repository; now, we need the updated RCS header
    "/
    (whatHappened isNil or:[whatHappened isEmpty]) ifTrue:[
        'SourceCodeManager [warning]: unexpected empty commit command output' errorPrintCR.
    ] ifFalse:[
        whatHappened := whatHappened asCollectionOfLines asStringCollection.
        idx := whatHappened indexOfLineStartingWith:'initial revision:'.
        idx ~~ 0 ifTrue:[
            startIdx := 18
        ] ifFalse:[
            idx := whatHappened indexOfLineStartingWith:'new revision:'.
            idx ~~ 0 ifTrue:[
                'SourceCodeManager [warning]: container existed before' errorPrintCR.
                startIdx := 14.
            ] ifFalse:[
                'SourceCodeManager [warning]: unexpected commit command output (no new-revision info)' errorPrintCR.
            ]
        ].
        idx ~~ 0 ifTrue:[
            l := whatHappened at:idx.
            newRevision := (l copyFrom:startIdx to:(l indexOf:$; startingAt:startIdx)-1) withoutSpaces.
        ]
    ].

    "/
    "/ patch the classes revisionInfo (but keep binaryRevision unchanged)
    "/ this makes everyone here believe, that the incore version of the class is based upon
    "/ the newly checked in version.
    "/ (however, the binaryRevision must remain as it is - we will need it to fetch the sourceCode
    "/  correctly for all unchanged methodss)
    "/

    newRevision notNil ifTrue:[
        changeLog := self 
                        revisionLogOf:aClass 
                        fromRevision:newRevision 
                        toRevision:newRevision
                        fileName:fileName
                        directory:packageDir
                        module:moduleDir.

        (changeLog isNil or:[(changeLog at:#revisions) size ~~ 1]) ifTrue:[
            'SourceCodeManager [error]: failed to update revisionString (no log)' errorPrintCR
        ] ifFalse:[
            entry := (changeLog at:#revisions) first.
            newString := self revisionStringFromLog:changeLog entry:entry forClass:aClass.
            aClass updateVersionMethodFor:newString.
            ('SourceCodeManager [info]: updated revisionString to:',newString) infoPrintCR
        ]
    ].

    cmdOut remove.
    tempdir recursiveRemove.

    Class addChangeRecordForClassCheckIn:aClass.
    ^ true

    "Created: 9.12.1995 / 19:13:37 / cg"
    "Modified: 23.4.1996 / 14:01:37 / stefan"
    "Modified: 14.2.1997 / 20:42:39 / cg"
!

createModule:moduleName
    "create a module directory"

    |tempdir moduleDir cmdOut cmd|

    "/
    "/ first, create a temporary work tree
    "/
    tempdir := self createTempDirectory:nil forModule:moduleName.
    tempdir isNil ifTrue:[^ false].

    cmdOut := Filename newTemporary.
    cmdOut exists ifTrue:[
        cmdOut remove.
    ].

    "/
    "/ create the module directory 
    "/
    moduleDir := tempdir construct:moduleName.
    moduleDir exists ifFalse:[
        'SourceCodeManager [error]: failed to create: ' errorPrint.  moduleDir name errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false.
    ].     

    "/
    "/ then import it into CVS
    "/

    cmd := 'import -m "created from browser" ' , '.' , ' initial initial > ' , cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut remove.
        tempdir recursiveRemove.
        ^ false.
    ].

    'SourceCodeManager [info]: created new module: ' infoPrint. moduleName infoPrintCR.
    tempdir recursiveRemove.
    ^ true

    "Created: 9.12.1995 / 19:53:51 / cg"
    "Modified: 23.4.1996 / 13:59:07 / stefan"
    "Modified: 10.1.1997 / 17:31:37 / cg"
!

createModule:moduleDirName package:packageDirName
    "create a package directory"

    |tempdir cmdOut cmd packageDir moduleDir where|

    "/
    "/ first, create a temporary work tree
    "/
    tempdir := self createTempDirectory:nil forModule:moduleDirName.
    tempdir isNil ifTrue:[^ false].

    cmdOut := Filename newTemporary.
    cmdOut exists ifTrue:[
        cmdOut remove.
    ].

    "/
    "/ first checkout that module directory (to let CVS create its files)
    "/

    cmd := 'checkout -l ' , moduleDirName , ' > ' , cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut remove.
        tempdir recursiveRemove.
        ^ false.
    ].

    moduleDir := tempdir construct:moduleDirName.
    moduleDir isDirectory ifFalse:[
        'SourceCodeManager [error]: command failed: ' errorPrint. cmd errorPrintCR.

        cmdOut remove.
        tempdir recursiveRemove.
        ^ false.
    ].
        
    "/
    "/ create the package directory there
    "/
    packageDir := moduleDir construct:packageDirName.
    OperatingSystem errorSignal catch:[
        packageDir recursiveMakeDirectory.
    ].
    packageDir isDirectory ifFalse:[
        cmdOut remove.
        tempdir recursiveRemove.
        ('SourceCodeManager [error]: cannot create new directory: ' , packageDir pathName) errorPrintCR.
        ^ false.
    ].

"/ old:
"/    "/
"/    "/ let cvs add this directory ...
"/    "/
"/    cmd := 'add ' , packageDirName , ' > ' , cmdOut name.
"/    (self executeCommand:cmd inDirectory:moduleDir name) ifFalse:[
"/        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.
"/
"/        cmdOut remove.
"/        tempdir recursiveRemove.
"/        ('SourceCodeManager [error]: cannot add ' , packageDirName) errorPrintCR.
"/        ^ false.
"/    ].

"/ new:

    "/
    "/ let cvs add all directories along the new packageDirPath ...
    "/
    where := moduleDir.
    (Filename components:packageDirName) do:[:aDir |
        cmd := 'add ' , aDir , ' > ' , cmdOut name.
        (self executeCommand:cmd inDirectory:where name) ifFalse:[
            'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

            cmdOut remove.
            tempdir recursiveRemove.
            ('SourceCodeManager [error]: cannot add ' , packageDirName) errorPrintCR.
            ^ false.
        ].
        where := where construct:aDir.
    ].

    cmdOut remove.

    self releaseAndRemove:tempdir outputTo:nil. 

    ^ true

    "Created: 9.12.1995 / 20:09:33 / cg"
    "Modified: 23.4.1996 / 14:28:43 / stefan"
    "Modified: 26.3.1997 / 18:15:49 / cg"
!

initialRevisionStringFor:aClass inModule:moduleDir package:packageDir container:fileName
    "return a string usable as initial revision string"

    ^ self
        initialRCSRevisionStringFor:aClass 
        in:(CVSRoot , '/' , moduleDir , '/' , packageDir)
        container:fileName

    "Modified: 14.2.1997 / 21:00:24 / cg"
!

removeContainerFor:aClass inModule:moduleDir package:packageDir container:fileName
    "remove a container"

    |fullName tempdir checkoutName cmdOut cmd tempFile whatHappened|

    "/
    "/ first, create a temporary work tree
    "/
    tempdir := self createTempDirectory:packageDir forModule:moduleDir.
    tempdir isNil ifTrue:[^ false].

    cmdOut := Filename newTemporary.
    cmdOut exists ifTrue:[
        cmdOut remove.
    ].

    fullName := moduleDir , '/' , packageDir , '/' , fileName.

    "/
    "/ first, check out everything there - this creates the CVS helpfiles
    "/ required later.
    "/
    checkoutName :=  moduleDir , '/' , packageDir.

    self activityNotification:'checking for directory ' , checkoutName.

    cmd := 'checkout -l ' , checkoutName , ' > ' , cmdOut name.
    (self executeCommand:cmd inDirectory:tempdir name) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut remove.
        tempdir recursiveRemove.
        ('SourceCodeManager [error]: cannot checkout ' , checkoutName) errorPrintCR.
        ^ false.
    ].

    "/
    "/ and remove it to the repository
    "/
    self activityNotification:'removing ' , fileName.

    "/
    "/ check presence of source there
    "/
    tempFile := (tempdir name , '/' , checkoutName , '/' , fileName) asFilename.
    tempFile exists ifFalse:[
        'SourceCodeManager [error]: temporary fileout failed' errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false
    ].

    tempFile renameTo:(tempFile pathName , '.sav').

    cmd := 'remove ' , fileName , ' > ' , cmdOut name.
    (self executeCommand:cmd inDirectory:(tempdir name , '/' , checkoutName)) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut remove.
        tempdir recursiveRemove.
        ('SourceCodeManager [error]: cannot remove ' , checkoutName) errorPrintCR.
        ^ false.
    ].

    cmdOut remove.

    "/
    "/ commit
    "/
    self activityNotification:'comitting removeal of ' , fileName.

    cmd := 'commit -m "removed container" -l ' , fileName , ' 2> ' , cmdOut name.
    (self executeCommand:cmd inDirectory:(tempdir name , '/' , checkoutName)) ifFalse:[
        'SourceCodeManager [error]: failed to execute: ' errorPrint. cmd errorPrintCR.

        cmdOut fileSize > 0 ifTrue:[
            whatHappened := cmdOut contentsOfEntireFile asString.
        ] ifFalse:[
            whatHappened := '<< no message >>'
        ].
        self warn:'The following problem was reported by cvs:

' , whatHappened , '

The container has NOT been removed into the repository.'.

       'SourceCodeManager [error]: cannot remove container' errorPrintCR.
        cmdOut remove.
        tempdir recursiveRemove.
        ^ false.
    ].
    whatHappened := cmdOut contentsOfEntireFile asString.

    "/
    "/ release the temporary tree towards cvs
    "/
    self releaseAndRemove:tempdir outputTo:nil. 

    "/
    "/ patch the classes revisionInfo (but keep binaryRevision unchanged)
    "/ this makes everyone here believe, that the incore version of the class is based upon
    "/ some container-less class.
    "/ (however, the binaryRevision must remain as it is - we will need it to fetch the sourceCode
    "/  correctly for all unchanged methodss)
    "/

    aClass updateVersionMethodFor:'no longer in repository'.
    Class addChangeRecordForClassContainerRemove:aClass.

    tempdir recursiveRemove.
    cmdOut remove.
    ^ true

    "Modified: 10.1.1997 / 17:34:07 / cg"
!

revisionInfoFromString:aString
    "{ Pragma: +optSpace }"

    "return a dictionary filled with revision info.
     This extracts the relevant info from aString."

    ^ self revisionInfoFromRCSString:aString

    "
     self revisionInfoFromString:'$Revision: 1.127 $'
     self revisionInfoFromString:(Array version)   
    "

    "Modified: 29.1.1997 / 19:00:35 / cg"
!

revisionLogOf:cls fromRevision:rev1 toRevision:rev2 fileName:classFileName directory:packageDir module:moduleDir
    "return info about the repository container and
     (part of) the revisionlog as a collection of revision entries.
     Return nil on failure.

     The returned information is a structure (IdentityDictionary)
     filled with:
            #container          -> the RCS container file name 
            #filename           -> the actual source file name
            #newestRevision     -> the revisionString of the newest revision
            #numberOfRevisions  -> the number of revisions in the container
            #revisions          -> collection of per-revision info (see below)

            rev1 / rev2 specify from which revisions a logEntry is wanted:
             -If rev1 is nil, the first revision is the initial revision
              otherwise, the log starts with that revision.
             -If rev2 is nil, the last revision is the newest revision
              otherwise, the log ends with that revision.

             -If both are nil, all logEntries are extracted.
             -If both are 0 (not nil), no logEntries are extracted (i.e. only the header).

            per revision info consists of one record per revision:

              #revision              -> the revision string
              #author                -> who checked that revision into the repository
              #date                  -> when was it checked in
              #state                 -> the RCS state
              #numberOfChangedLines  -> the number of changed line w.r.t the previous

            revisions are ordered newest first 
            (i.e. the last entry is for the initial revision; the first for the most recent one)
        "

    |tempDir cmd fullName modulePath inStream inHeaderInfo atEnd line revArg revLine1 revLine2 idx
     info record revisionRecords s headerOnly msg|

    CVSRoot isNil ifTrue:[
        'SourceCodeManager [info]: CVSROOT not set' infoPrintCR.
        ^ nil
    ].

    modulePath :=  moduleDir , '/' , packageDir. 
    fullName :=  modulePath , '/' , classFileName.
    tempDir := self createTempDirectory:nil forModule:nil.
    [
        self createEntryFor:fullName 
             in:(tempDir construct:modulePath) 
             revision:'1.1' 
             date:'dummy' 
             special:''
             overwrite:false.

        revArg := ''.
        headerOnly := false.
        (rev1 notNil or:[rev2 notNil]) ifTrue:[
            (rev1 == 0 and:[rev2 == 0]) ifTrue:[
                revArg := '-h'.
                headerOnly := true.
            ] ifFalse:[
                revArg := '-r'.
                rev1 notNil ifTrue:[
                    revArg := revArg , rev1
                ].
                revArg := revArg , ':'.
                rev2 notNil ifTrue:[
                    revArg := revArg , rev2
                ].
            ]
        ].

        headerOnly ifTrue:[
            msg := 'fetching revision info '
        ] ifFalse:[
            msg := 'reading revision log '
        ].
        cls isNil ifTrue:[
            msg := msg , 'in ' , fullName.
        ] ifFalse:[
            msg := msg , 'of ', cls name.
        ].
        self activityNotification:msg.

        cmd := 'cd ', tempDir name, '; cvs -d ', CVSRoot, ' log ' , revArg , ' ' , fullName.
    "/ cmd printNL.
        inStream := PipeStream readingFrom:cmd.
        inStream isNil ifTrue:[
            ('SourceCodeManager [error]: cannot open pipe to cvs log ', fullName) errorPrintCR.
            ^ nil
        ].

        "/
        "/ read the commands pipe output and extract the container info
        "/
        info := IdentityDictionary new.
        inHeaderInfo := true.
        [inHeaderInfo and:[inStream atEnd not]] whileTrue:[
            line:= inStream nextLine.
            line notNil ifTrue:[
                |gotIt|

                gotIt := false.
                #('RCS file:' #container
                  'Working file:' #filename
                  'head:' #newestRevision
                  'total revisions:' #numberOfRevisions
                 ) pairWiseDo:[:word :key |
                    gotIt ifFalse:[
                        s := line restAfter:word withoutSeparators:true.
                        s notNil ifTrue:[info at:key put:s. gotIt := true].                        
                    ]
                ].
                gotIt ifFalse:[
                    (line startsWith:'description:') ifTrue:[inHeaderInfo := false].
                ]
            ]
        ].
        inStream nextLine. "/ skip separating line after description.

        info isEmpty ifTrue:[
            ('SourceCodeManager [warning]: no log for ', fullName) errorPrintCR.
            ^ nil
        ].

        "/ strip selected revisions from the total-revisions entry
        s := info at:#numberOfRevisions.
        (idx := s indexOf:$;) ~~ 0 ifTrue:[
            info at:#numberOfRevisions put:(Integer readFrom:(s copyTo:idx - 1))
        ].

        headerOnly ifFalse:[
            "/
            "/ continue to read the commands pipe output 
            "/ and extract revision info records
            "/
            revisionRecords := OrderedCollection new:(info at:#numberOfRevisions).
            info at:#revisions put:revisionRecords.

            atEnd := false.
            [atEnd or:[inStream atEnd]] whileFalse:[
                revLine1 := inStream nextLine.
                [revLine1 notNil and:[(revLine1 startsWith:'revision ') not]]
                    whileTrue:[revLine1 := inStream nextLine.].
                revLine2 := inStream nextLine.
                (revLine1 notNil and:[revLine2 notNil]) ifTrue:[
                    record := IdentityDictionary new.
                    record at:#revision put:(revLine1 asCollectionOfWords at:2).
                    "/ decompose date/author/state etc.
                    (revLine2 asCollectionOfSubstringsSeparatedBy:$;) do:[:info |
                        |subEntry|
                        subEntry := info withoutSeparators.
                        #('date:' #date
                          'author:' #author 
                          'state:' #state 
                          'lines:' #numberOfChangedLines
                         ) pairWiseDo:[:word :key |
                            s := subEntry restAfter:word withoutSeparators:true.
                            s notNil ifTrue:[record at:key put:s.].                        
                        ].
                    ].

                    "first revision does not hav a 'lines:' entry"
                    (record includesKey:#numberOfChangedLines) ifFalse:[
                        record at:#numberOfChangedLines put:''
                    ].

                    s := nil.
                    line := inStream nextLine.
                    [atEnd or:[line isNil or:[line startsWith:'--------']]] whileFalse:[
                        (line startsWith:'==========') ifTrue:[
                            atEnd := true.
                        ] ifFalse:[
                            (line withoutSpaces = '.') ifTrue:[
                                line := '*** empty log message ***'
                            ].
                            s isNil ifTrue:[
                                s := line
                            ] ifFalse:[
                                s := s , Character cr asString , line.
                            ].
                            line := inStream nextLine.
                        ]
                    ].
                    record at:#logMessage put:s.
                    revisionRecords add:record.
                ]
            ].
        ].
    ] valueNowOrOnUnwindDo:[
        inStream notNil ifTrue:[inStream close].
        tempDir recursiveRemove
    ].
    ^ info

    "
     SourceCodeManager revisionLogOf:Array 
     SourceCodeManager revisionLogOf:Array fromRevision:'1.40' toRevision:'1.43' 
     SourceCodeManager revisionLogOf:Array fromRevision:'1.40' toRevision:nil  
     SourceCodeManager revisionLogOf:Array fromRevision:nil toRevision:'1.3'  
     SourceCodeManager revisionLogOf:Array fromRevision:nil toRevision:nil  
     SourceCodeManager revisionLogOf:Array fromRevision:0 toRevision:0  
    "

    "Created: 16.11.1995 / 13:25:30 / cg"
    "Modified: 10.1.1997 / 15:31:51 / cg"
    "Modified: 29.1.1997 / 16:51:30 / stefan"
!

writeHistoryLogSince:timeGoal filterSTSources:filter filterUser:userFilter to:aStream
    "send a full historyLog to some stream"

    |cmd inStream line words recordType fileName user date time rev pkgDir 
     clsName cls clsRev goalString prevUser|

    CVSRoot isNil ifTrue:[
        'SourceCodeManager [info]: CVSROOT not set' infoPrintCR.
        ^ nil
    ].

    goalString := ''.
    cmd := 'cvs -d ' , CVSRoot , ' history -xWARM -a'.

    (timeGoal notNil and:[timeGoal notEmpty]) ifTrue:[
        cmd := cmd , ' -D "' , timeGoal , '"'.
        goalString := ' since ' , timeGoal.
    ].

"/ cmd printNL.
    inStream := PipeStream readingFrom:cmd.
    inStream isNil ifTrue:[
        ('SourceCodeManager [error]: cannot open pipe to ''cvs history''') errorPrintCR.
        ^ nil
    ].

    aStream nextPutLine:'**** source repository history' , goalString , ' ****'.
    aStream cr; cr.
    aStream nextPutLine:'  date  time        user                     file    rev        module/package'.

    "/
    "/ read the commands pipe output and reformat the lines
    "/
    [inStream atEnd not] whileTrue:[
        inStream readWait.
        line := inStream nextLine.
        line notNil ifTrue:[
            "/ cvs history line is of the following format:
            "/ [Ma] date time +xxx user rev name module/package ....
            words := line asCollectionOfWords.
            words size >= 8 ifFalse:[
                "/ something like 'No records selected' ...
                aStream nextPutLine:line
            ] ifTrue:[
                recordType := words at:1.
                fileName := words at:7.
                user := words at:5.

                (filter not 
                or:[fileName endsWith:'.st']) ifTrue:[
                    (userFilter isNil
                    or:[userFilter = user]) ifTrue:[

                        date := words at:2.
                        time := words at:3.
                        rev := words at:6.
                        pkgDir := words at:8.

                        prevUser ~= user ifTrue:[
                            aStream cr.
                            prevUser := user.
                        ].
                        recordType = 'M' ifTrue:[
                            recordType := ' '
                        ] ifFalse:[
                            recordType = 'A' ifTrue:[
                                recordType := '+'
                            ] ifFalse:[
                                (recordType = 'R' or:[recordType = 'W']) ifTrue:[
                                    recordType := '-'
                                ]
                            ]
                        ].

                        aStream nextPutAll:recordType; space; 
                                nextPutAll:date; space; nextPutAll:time; space;
                                nextPutAll:(user leftPaddedTo:10); space;             
                                nextPutAll:(fileName leftPaddedTo:24); space;                 
                                nextPutAll:(rev decimalPaddedTo:8 and:3 at:$. withLeft:(Character space) right:nil); tab;              
                                nextPutAll:(pkgDir paddedTo:20).

                        "/
                        "/ for your convenience:
                        "/  check what the actual version is in the image
                        "/
                        clsName := Smalltalk classNameForFile:fileName.
                        clsName notNil ifTrue:[
                            cls := Smalltalk classNamed:clsName.
                            (cls notNil and:[(clsRev := cls revision) notNil]) ifTrue:[
                                rev ~= clsRev ifTrue:[
                                    aStream nextPutAll:('current: ' , clsRev)
                                ]
                            ] ifFalse:[
                                cls isNil ifTrue:[
                                    aStream nextPutAll:('current: ** none **')
                                ] ifFalse:[
                                    cls isLoaded not ifTrue:[
                                        aStream nextPutAll:('current: not loaded')
                                    ] ifFalse:[
                                        aStream nextPutAll:('current: ** no revision info **')
                                    ]    
                                ]
                            ]
                        ].
                        aStream cr                              
                    ]
                ]
            ]
        ]
    ].
    inStream close.

    "Created: 13.12.1995 / 10:28:27 / cg"
    "Modified: 10.1.1997 / 15:35:31 / cg"
! !

!CVSSourceCodeManager class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic3/Attic/CVSSCMgr.st,v 1.127 1997-03-26 17:34:26 cg Exp $ [ce]'
! !
CVSSourceCodeManager initialize!
